<?php

require_once("ews/ExchangeServices.php");
require_once("pbx/pbx.php");
require_once("include.php");

tl::register("exchangeuser", dirname(__FILE__) . "\\lang");

$out = main_system("ps auxwww | grep -v \"grep\" | egrep -e \".*/bin/sh.*php5.*innovaphone-exchange/exchange.php.*\"", $ret);
$lines = explode("\n", trim($out, "\n"));
if(count($lines) > 1) {    // already running?
    var_dump($lines);
    MainBackup::doLog("Script already running, exiting...");
    exit;
}

class ExchangeAvailability
{
    /**
     * Holds the ExchangeServices class
     *
     * @var ExchangeServices 
     *
     */    
    var $ews = null;
    /**
     * holds the ExchangeServices version
     *
     * @var EWS_RequestServerVersion 
     *
     */    
    var $version = null;
    
    /**
     * holds the pbx soap class
     *
     * @var pbx
     *
     */ 
    var $pbxSoap = null;
    
    /**
     *
     * @var ExchangeServer
     */
    var $server = null;
    
    var $users = array();
    
    var $linuxUrl = "";
    
    const maxPbxUsers                   = 50;       // retrieve maxPbxUsers at once from PBX
    const maxEwsUsers                   = 5;        // retrieve calendaritems for maxEwsUsers at once from EWS
    const availabilityThreshold         = 12;       // threshold for availability timespan for refresh purposes in hours
    const pushStatusFrequencyThreshold  = 2;        // threshold for the push status frequency, which is a multiplier of the original status frequency
    
    const calStatusUnset = 0;
    const calStatusNotify = 1;
    const calStatusStartSet = 2;
    const calStatusEndSet = 3;
    
    var $trace = true;
    
    public function ExchangeAvailability($id) {
        $this->server = new ExchangeServer();
        if($this->server->read($id) === false) {
            $this->debug("ERROR: server with id $id couldn't be read");
            $this->server = null;
            return;
        }
        global $translator;
        $translator->change($this->server->lang);
        $this->trace = $this->server->trace;
        $this->linuxUrl = "http";
        if($this->server->ews->ssl) {
            $this->linuxUrl .= "s";
        }
        $this->linuxUrl .= "://";
        if($this->server->ews->linuxIp != "") {
            $this->linuxUrl .= $this->server->ews->linuxIp;
            if($this->server->ews->linuxPort != "") {
                $this->linuxUrl .= ":".$this->server->ews->linuxPort;
            }
        }
        else {
            $this->linuxUrl .= get_dhcp_address();
        }
        $this->linuxUrl .= "/apps/innovaphone-exchange/subscriptions.php";
        $this->debug("Linux URL: {$this->linuxUrl}");
        try
        {
            $this->pbxInit();
            $this->debug("PBX initialized");
            $this->exchangeServiceInit();
            $this->debug("EWS initialized");
        }
        catch(PbxException $e) {
            $this->handlePbxException($e);
        }
    }
    
    public function run() {
        if($this->server === null) return;
        try
        {
            $this->fetchUsersFromDb();
            $this->debug("Users fetched");
            $this->findUsers();
            $this->debug("Users matched");
            $this->retrieveCalendarItems();
            $this->debug("Calendar items retrieved");
            $this->updatePbxPresence();
            $this->debug("PBX presence updated");
            $this->updateLastRun();
        }
        catch(PbxException $e) {
            $this->handlePbxException($e);
        }
    }
    
    public function handlePbxException(PbxException &$e) {
        $this->debug("ERROR: ".$e->getMessage().": {$e->details} ({$e->errorCode})");
        send_alarm($this->server->name.": ".$e->getMessage(), $e->details, $e->errorCode, false, ExchangeMethods::getAlarmServer(), "", false, "Exchange");
    }
    
    static public function debugPrefix() {
        $time = gettimeofday();
        return date(MainInputPage::dateFormat, $time["sec"]).".".substr($time["usec"], 0, 3).": ";
    }
    
    private function debug($msg) {
        if($this->trace) {
            echo self::debugPrefix()."{$this->server->name}: $msg\n";
        }
    }
    
    private function getMailBox($mail) {
        $mailBox = new EWS_MailboxData();

        // Identify the user mailbox to review for free/busy data.
        $emailAddress = new EWS_EmailAddress();
        $emailAddress->Address = $mail;
        $emailAddress->Name = "";

        $mailBox->Email = $emailAddress;
        $mailBox->ExcludeConflicts = false;
        $mailBox->AttendeeType = EWS_MailboxData::AttendeeType_Organizer;
        return $mailBox;
    }
    
    private function exchangeServiceInit() {
        $this->ews = $this->server->ewsSoap();
        if($this->ews === false) {
            throw new PbxException("Exchange Web Service connection failed", $this->server->error, Alarms::alExchangeEwsInit);
        }
        $this->version = new EWS_RequestServerVersion();
        $this->version->Version = $this->server->ews->version;
        clear_alarm(Alarms::alExchangeEwsInit, ExchangeMethods::getAlarmServer(), "Exchange");
    }
    
    private function fetchUsersFromDb() {
        $exchangeDb = get_exchange_db();
        $query = "SELECT guid, id, h323, emails, primary_smtp, subscription_id,
                        date_part('epoch', subscription_status) AS subscription_status, date_part('epoch', availability_status) AS availability_status 
                    FROM users WHERE server_id={$this->server->id}";
        $result = $exchangeDb->run($query);
        while(($row = pg_fetch_object($result))) {
            $this->users[$row->guid] = new PbxUser($row->guid, $row->id, $row->h323, $row->emails, $row->primary_smtp, $row->subscription_id, intval($row->subscription_status), intval($row->availability_status));
        }
    }
    
    private function pbxInit() {
        $this->pbxSoap = $this->server->pbxSoap();
        if($this->pbxSoap === false) {
            throw new PbxException("PBX SOAP connection failed", $this->server->error, Alarms::alExchangeSoapInit);
        }
        clear_alarm(Alarms::alExchangeSoapInit, ExchangeMethods::getAlarmServer(), "Exchange");
    }
    
    private function findUsers() {
        $exchangeDb = get_exchange_db();
        $version = null;
        try
        {
            $version = $this->pbxSoap->Version();
        }
        catch(Exception $e) {
            throw new PbxException("Couldn't retrieve PBX version", $e->getMessage(), Alarms::alExchangeSoapVersion);
        }
        clear_alarm(Alarms::alExchangeSoapVersion, ExchangeMethods::getAlarmServer(), "Exchange");
        if(!isset($version["WSDLVersion"]) || $version["WSDLVersion"] < 1000) {
            throw new PbxException("Wrong PBX WSDL version {$version['WSDLVersion']}", "You need a PBX with Firmware >= 10.00.", Alarms::alExchangeSoapWrongVersion);
        }
        clear_alarm(Alarms::alExchangeSoapWrongVersion, ExchangeMethods::getAlarmServer(), "Exchange");
        $i = 0;
        $next = false;
        $lastCn = "";
        
        do {
            $i = 0;
            $users = array();
            try
            {
                $this->debug("Find users");
                $users = $this->pbxSoap->FindUser(true, true, true, true, $lastCn, null, null, self::maxPbxUsers, $next);
                $this->debug("Users found: ".count($users));
            } 
            catch(Exception $e) {
                throw new PbxException("PBX FindUser failed", $e->getMessage(), Alarms::alExchangeSoapFindUser);
            }
            clear_alarm(Alarms::alExchangeSoapFindUser, ExchangeMethods::getAlarmServer(), "Exchange");
            $next = true;
            foreach($users as /** @var PBX_UserInfo $user */ $user) {
                $i++;
                $lastCn = $user->cn;
                $user->type = strtolower($user->type);
                if(($user->type != "ep" && $user->type != "executive") || $user->h323 == "") continue;
                $emails = array();
                $domain = ((isset($user->domain) && $user->domain != "") ? $user->domain : null);
                foreach($user->email as $email) {
                    if(strpos($email, "@") !== false) {
                        $emails[$email] = $email;
                    }
                    elseif(isset($domain)) {
                        $mail = $email."@".$domain;
                        $emails[$mail] = $mail;
                    }
                }
                if($user->h323email && $user->h323 != "" && isset($domain)) {
                    $mail = $user->h323."@".$domain;
                    $emails[$mail] = $mail;
                }
                unset($email);
                foreach($emails as $index => $email) {
                    if(!valid_email($email) || !$this->server->pbx->checkDomain($email)) {
                        $this->debug("ignoring email $email");
                        unset($emails[$index]);
                    }
                }
                
                if(!count($emails)) continue;
                
                $emailStr = implode(";", $emails);
                foreach($this->users as $guid => &$pbxUser) {
                    if($guid == $user->guid) {
                        $pbxUser->exists = true;
                        if($pbxUser->h323 != $user->h323 || $pbxUser->emails != $emailStr || $pbxUser->smtp == "") {
                            $error = "";
                            $smtp = $this->getPrimarySmtp($emails, $error);
                            if($smtp != $pbxUser->smtp) {
                                $this->unsubscribe($pbxUser);
                            }
                            if($smtp == "") {
                                $this->debug("No SMTP address for $emailStr found: $error");
                            }
                            $this->updateUser($pbxUser, $user->h323, $emailStr, $smtp);
                        }
                        break;
                    }
                    elseif($pbxUser->h323 == $user->h323) {
                        $pbxUser->exists = true;
                        $oldGuid = $pbxUser->guid;
                        $this->updateUserGuid($pbxUser, $user->guid);
                        $this->debug("User guid of {$pbxUser->h323} changed");
                        break;
                    }
                }
                if(!isset($this->users[$user->guid])) {
                    $this->debug("New user {$user->h323}");
                    $newUser = new PbxUser($user->guid, 0, $user->h323, $emailStr, "", "", 0, 0);
                    $this->addUser($newUser, $emails);
                }
            }
        } while($i == self::maxPbxUsers);
    }
    
    private function retrieveCalendarItems() {
        $delIds = "";
        $availabilityRequests = array();
        foreach($this->users as /* @var $pbxUser PbxUser */ &$pbxUser) {
            $time = time();
            $availabilityTimespan = (($this->server->ews->availabilityTimespan - self::availabilityThreshold) * 60 * 60);
            if(!$pbxUser->exists) {
                $this->debug("Deleting obsolete user {$pbxUser->h323}"); 
                $this->unsubscribe($pbxUser);
                $delIds .= $pbxUser->id.",";
                unset($this->users[$pbxUser->guid]);
            }
            else if($pbxUser->smtp != "") {
                if(($time - $pbxUser->subscriptionStatus) > (($this->server->ews->statusFrequency * self::pushStatusFrequencyThreshold) * 60)) {
                    if($pbxUser->subscriptionId != "") {    // dead subscription
                        $this->debug("Dead subscription for user {$pbxUser->h323}"); 
                        $this->unsubscribe($pbxUser);
                    }
                    $this->subscribe($pbxUser);
                    if($pbxUser->subscriptionId != "") {
                        $availabilityRequests[] = $pbxUser;
                    }
                }
                else if ($pbxUser->subscriptionId != "" && (($time - $pbxUser->availabilityStatus) > $availabilityTimespan)) {  // request next calendar entries
                    $this->debug("Refresh calendar entries for user {$pbxUser->h323}"); 
                    $availabilityRequests[] = $pbxUser;
                }
            }
        }
        $delIds = rtrim($delIds, ",");
        if($delIds != "") { // delete users with their cal entries
            $this->deleteUsers($delIds);
        }
        
        $mailCount = 0;
        $timeOffset = date_offset_get(new DateTime(date_default_timezone_get()));
        /** @var PbxUser */$pbxUser = reset($availabilityRequests);
        do
        {
            $mailboxes = array();
            $mailCount = 0;
            $pbxUsers = array();
            while($pbxUser !== false && $mailCount < self::maxEwsUsers) {
                $mailboxes[] = $this->getMailBox($pbxUser->smtp);
                $pbxUsers[] = $pbxUser;
                $pbxUser = next($availabilityRequests);
                $mailCount++;
            }
            if(count($mailboxes) === 0) break;
            
            // Identify the time to compare free/busy information.
            $duration = new EWS_Duration();
            $duration->StartTime = gmdate('Y-m-d\TH:i:s.0000000\Z', gmmktime());
            $duration->EndTime = gmdate('Y-m-d\TH:i:s.0000000\Z', gmmktime() + $this->server->ews->availabilityTimespan * 60 * 60 );

            // Identify the options for comparing free/busy information.
            $fbViewOptions = new EWS_FreeBusyViewOptionsType();
            $fbViewOptions->TimeWindow = $duration;
            $fbViewOptions->RequestedView = EWS_FreeBusyViewOptionsType::RequestedView_Detailed;
            
            // Make the request.
            $request = new EWS_GetUserAvailabilityRequestType();

            // Set the time zone of the request (UTC!)
            $request->TimeZone = new EWS_SerializableTimeZone();
            $request->TimeZone->Bias = 0;
            $request->TimeZone->StandardTime = new EWS_SerializableTimeZoneTime();
            $request->TimeZone->StandardTime->Bias = 0;
            $request->TimeZone->StandardTime->DayOfWeek = EWS_SerializableTimeZoneTime::DayOfWeek_Sunday;
            $request->TimeZone->StandardTime->DayOrder = 1;
            $request->TimeZone->StandardTime->Month = 1;
            $request->TimeZone->StandardTime->Time = "00:00:00";
            $request->TimeZone->DaylightTime = new EWS_SerializableTimeZoneTime();
            $request->TimeZone->DaylightTime->Bias = 0;
            $request->TimeZone->DaylightTime->DayOfWeek = EWS_SerializableTimeZoneTime::DayOfWeek_Sunday;
            $request->TimeZone->DaylightTime->DayOrder = 5;
            $request->TimeZone->DaylightTime->Month = 12;
            $request->TimeZone->DaylightTime->Time = "23:59:59";

            // Add the mailboxes to the request.
            $request->MailboxDataArray = $mailboxes;

            // Add the view options to the request.
            $request->FreeBusyViewOptions = $fbViewOptions;

            /** @var EWS_GetUserAvailabilityResponseType*/$result = null;
            try
            {
                $result = $this->ews->GetUserAvailability($request, null, null, $this->version);
            }
            catch(Exception $e) {
                throw new PbxException("Getting user availability failed", $e->getMessage(), Alarms::alExchangeGetAvailability);
            }
            clear_alarm(Alarms::alExchangeGetAvailability, ExchangeMethods::getAlarmServer(), "Exchange");
            if($result->FreeBusyResponseArray != null && $result->FreeBusyResponseArray->FreeBusyResponse != null) {
                $freeBusyResponses = $result->FreeBusyResponseArray->FreeBusyResponse;
                if(!is_array($freeBusyResponses)) {
                    $freeBusyResponses = array($freeBusyResponses);
                }
                foreach($freeBusyResponses as $index => /** @var EWS_FreeBusyResponseType */ $freeBusyResponse) {
                    $calendarEvents = array();
                    if($freeBusyResponse->ResponseMessage->ResponseClass == EWS_ResponseMessageType::ResponseClass_Success) {
                        if($freeBusyResponse->FreeBusyView && $freeBusyResponse->FreeBusyView->CalendarEventArray && $freeBusyResponse->FreeBusyView->CalendarEventArray->CalendarEvent) {
                            $calendarEvents = $freeBusyResponse->FreeBusyView->CalendarEventArray->CalendarEvent;
                            if(!is_array($calendarEvents)) {
                                $calendarEvents = array($calendarEvents);
                            }
                        }
                    }
                    else {
                         $this->debug("FreeBusyResponseClass user {$pbxUsers[$index]->h323}: {$freeBusyResponse->ResponseMessage->ResponseClass}");
                    }
                    //echo "calendar entries for ".$pbxUsers[$index]->h323.":<br>";
                    $this->debug("Found ".count($calendarEvents)." calendar entries for {$pbxUsers[$index]->h323}");
                    $this->updateCalendarItems($pbxUsers[$index], $calendarEvents, $timeOffset);
                }
            }
            else {  // no entries received, just update availability and deleted calendar items
                foreach($pbxUsers as $pbxUser) {
                    $this->updateCalendarItems($pbxUser, array(), 0);
                }
            }
        } while(true);
    }
    
    private function sortCalendarItems(EWS_CalendarEvent &$left, EWS_CalendarEvent &$right) {
        return strtotime($left->StartTime) - strtotime($right->StartTime);
    }
    
    private function deleteCalendarItems(PbxUser &$user) {
        $exchangeDb = get_exchange_db();
        $query = "DELETE FROM calendar_items WHERE user_id={$user->id}";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("Deleting user calendar items failed", pg_last_error($exchangeDb->db), Alarms::alExchangeDeleteCalendarItems);
        }
    }
    
    private function updateCalendarItems(PbxUser &$user, $calendarEvents, $timeOffset) {
        $exchangeDb = get_exchange_db();
        $exchangeDb->run("BEGIN");
        // $exchangeDb->run("LOCK calendar_items IN SHARE MODE"); // no lock needed, as there is no concurrent WRITE
        // retrieve all items with already set presence to maintain pbx status
        $query = "SELECT * "
                . "FROM calendar_items "
                . "WHERE user_id={$user->id} AND (pbx_status=".self::calStatusNotify." OR pbx_status=".self::calStatusStartSet.")";
        $result = $exchangeDb->run($query);
        $currentItems = array();
        while(($row = pg_fetch_object($result))) {
            if(isset($row->exchange_id)) {
                $currentItems[$row->exchange_id] = $row;
            }
        }
        // delete all current calendar_items, as there is no need to maintain old items or items further in the future
        $this->deleteCalendarItems($user);
        clear_alarm(Alarms::alExchangeDeleteCalendarItems, ExchangeMethods::getAlarmServer(), "Exchange");
        $validCalendarEvents = 0;
        uasort($calendarEvents, array($this, "sortCalendarItems"));
        $unsetExists = false;
        foreach($calendarEvents as /* @var $event EWS_CalendarEvent */ $event) {
            $busyType = $event->BusyType;
            $itemId = $event->CalendarEventDetails->ID;
            if($itemId == "") {
                $itemId = strtotime($event->StartTime).strtotime($event->EndTime);
            }
            if($event->BusyType == EWS_CalendarEvent::BusyType_Free) {
                // ignore, as the user has time for Free or forget to update the busy type for tentative
                continue;
            }
            else if($event->BusyType == EWS_CalendarEvent::BusyType_Tentative) {
                if($event->CalendarEventDetails->ID == "") {    // private entry
                    continue;
                }
                /* @var $msg EWS_ConvertIdResponseMessageType */$msg;
                try {
                    $request = new EWS_ConvertIdType();
                    $request->DestinationFormat = EWS_ConvertIdType::DestinationFormat_EwsId;
                    $request->SourceIds = new EWS_NonEmptyArrayOfAlternateIdsType();
                    $altId = new EWS_AlternateIdType();
                    $altId->Id = $event->CalendarEventDetails->ID;
                    $altId->Format = EWS_AlternateIdType::Format_HexEntryId;
                    $altId->Mailbox = $user->smtp;
                    $request->SourceIds->AlternateId[] = $altId;
                    $response = $this->ews->ConvertId($request, null, $this->version);
                    $msg = $response->ResponseMessages->ConvertIdResponseMessage;
                    if($msg->ResponseClass != EWS_ResponseMessageType::ResponseClass_Success) {
                        $this->debug("Converting calendar ID failed: {$msg->MessageText}");
                        continue;
                    }
                    elseif($msg->AlternateId->Id == "") {
                        $this->debug("Converting calendar ID failed: no ID returned");
                        continue;
                    }
                }
                catch (Exception $e) {
                    throw new PbxException("Converting calendar ID request failed", $e->getMessage(), Alarms::alExchangeConvertCalendarId);
                }
                clear_alarm(Alarms::alExchangeConvertCalendarId, ExchangeMethods::getAlarmServer(), "Exchange");
                try {
                    $request = new EWS_GetItemType();
                    $request->ItemShape = new EWS_ItemResponseShapeType();
                    $request->ItemShape->BaseShape = EWS_ItemResponseShapeType::BaseShape_AllProperties;
                    $calId = new EWS_ItemIdType();
                    $calId->Id = $msg->AlternateId->Id;
                    $request->ItemIds = new EWS_NonEmptyArrayOfBaseItemIdsType();
                    $request->ItemIds->ItemId[] = $calId;
                    $response = $this->ews->GetItem($request, null, null, $this->version);
                    /* @var $msg2 EWS_ItemInfoResponseMessageType */ $msg2 = $response->ResponseMessages->GetItemResponseMessage;
                    if($msg2->ResponseClass != EWS_ResponseMessageType::ResponseClass_Success) {
                        $this->debug("Get calendar item failed: {$msg2->MessageText}");
                        continue;
                    }
                    if(!is_object($msg2->Items->CalendarItem->Organizer->Mailbox)) {
                        $this->debug("Couldn't get calendar item organizer");
                        continue;
                    }
                    $organizer = $msg2->Items->CalendarItem->Organizer->Mailbox->EmailAddress;
                    $this->debug("Organizer of tentative calendar entry: $organizer");
                    if(strpos($this->server->ews->tentativeMails, $organizer) === false) {
                        continue;
                    }
                }
                catch(Exception $e) {
                    throw new PbxException("Getting calendar item failed", $e->getMessage(), Alarms::alExchangeGetItemCalendar);
                }
                clear_alarm(Alarms::alExchangeGetItemCalendar, ExchangeMethods::getAlarmServer(), "Exchange");
                $busyType = EWS_CalendarEvent::BusyType_OOF;    // override busy type
            }
            $validCalendarEvents++;
            $pbxStatus = self::calStatusUnset;
            $startTime = date("Y-m-d H:i:s", strtotime($event->StartTime) + $timeOffset);
            $endTime = date("Y-m-d H:i:s", strtotime($event->EndTime) + $timeOffset);
            $isMeeting = ($event->CalendarEventDetails->IsMeeting ? "t" : "f");
            $isPrivate = ($event->CalendarEventDetails->IsPrivate ? "t" : "f");
            if(!$unsetExists && isset($currentItems[$itemId])) {
                // check if calendar item has changed
                $item = $currentItems[$itemId];
                if($item->subject != $event->CalendarEventDetails->Subject ||
                    $item->time_start != $startTime ||
                    $item->time_end != $endTime ||
                    $item->location != $event->CalendarEventDetails->Location ||
                    $item->is_meeting != $isMeeting ||
                    $item->is_private != $isPrivate ||
                    $item->busy_type != $busyType) {
                    $pbxStatus = self::calStatusUnset;  // item has changed, so reset the PBX calendar presence
                }
                else {
                    $pbxStatus = $item->pbx_status;
                }
            }
            if($pbxStatus == self::calStatusUnset) {
                $unsetExists = true;
            }
            // calendar ID is empty for private items!
            $query = "INSERT INTO calendar_items (user_id, subject, time_start, time_end, location, is_meeting, is_private, busy_type, exchange_id, pbx_status) VALUES (";
            $query .= "{$user->id},";
            $query .= "'".pg_escape_string($event->CalendarEventDetails->Subject)."',";
            $query .= "'$startTime',";
            $query .= "'$endTime',";
            $query .= "'".pg_escape_string($event->CalendarEventDetails->Location)."',";
            $query .= "'$isMeeting',";
            $query .= "'$isPrivate',";
            $query .= "'".pg_escape_string($busyType)."',";
            $query .= "'".pg_escape_string($itemId)."',";
            $query .= "$pbxStatus";
            $query .= ")";
            if($exchangeDb->run($query) === false) {
                throw new PbxException("Inserting user calendar item failed", pg_last_error($exchangeDb->db), Alarms::alExchangeInsertCalendarItem);
            }
            clear_alarm(Alarms::alExchangeInsertCalendarItem, ExchangeMethods::getAlarmServer(), "Exchange");
        }
        if($validCalendarEvents == 0) {
            $this->setPresence(new PBX_Presence(), $user->guid, $user->h323);
        }
        $this->updateUserAvailability($user);
        $exchangeDb->run("COMMIT");
    }
    
    private function setPresence(PBX_Presence &$presence, $guid, $h323) {
        try
        {
            $presenceReturn = $this->pbxSoap->SetPresence($presence, false, PbxConfig::presenceContact, $guid, null);
            $this->debug("Presence update for {$h323}: ({$presence->note}, {$presence->activity}, {$presence->status})");
        }
        catch(Exception $e) {
            throw new PbxException("PBX presence update failed", $e->getMessage(), Alarms::alExchangePresenceUpdate);
        }
        clear_alarm(Alarms::alExchangePresenceUpdate, ExchangeMethods::getAlarmServer(), "Exchange");
    }
    
    private function updatePbxPresence() {
        $exchangeDb = get_exchange_db();
        // pbx_status=0 if no other calendar_item is already notified/set on the pbx and the end time is still greater than now
        // pbx_status=1 if start time is reached
        // pbx_status=2 if end time is reached
        // this query select rows partitioned by user_id and just the first row of all rows of a user (x.row_num=1)
        $query = "SELECT x.*, u.guid, u.h323 FROM 
                     (
	                    SELECT ROW_NUMBER() OVER (PARTITION BY user_id order by time_start ASC) AS row_num,
	                    ci.*
	                    from calendar_items ci
	                    where (ci.pbx_status=".self::calStatusUnset." AND ci.time_end >= now() AND NOT EXISTS (SELECT t.id FROM calendar_items t WHERE (t.pbx_status=1 OR t.pbx_status=2) AND t.user_id=ci.user_id)) OR
                                  (ci.pbx_status=".self::calStatusNotify." AND ci.time_start <= now()) OR
                                  (ci.pbx_status=".self::calStatusStartSet." AND ci.time_end <= now()) 
                     ) x
                    JOIN users u ON x.user_id=u.id
                    WHERE x.row_num = 1 AND u.server_id={$this->server->id}";
        $result = $exchangeDb->run($query);
        if($result === false) {
            throw new PbxException("Getting calendar items failed", pg_last_error($exchangeDb->db), Alarms::alExchangeGetCalendarItems);
        }
        clear_alarm(Alarms::alExchangeGetCalendarItems, ExchangeMethods::getAlarmServer(), "Exchange");
        $updateUnset = "";
        $updateNotfiy = "";
        $updateStart = "";
        while(($row = pg_fetch_object($result))) {
            $presence = new PBX_Presence();
            $presence->note = null;
            $presence->activity = null;
            $presence->status = null;
            switch($row->pbx_status) {
                case self::calStatusUnset:
                    if(strtotime($row->time_start) <= time()) { // this entry has already started
                        // do not break here, fall through to calStatusNotify!
                    }
                    else {                                      // entry not yet started, just inform the PBX about it
                        $updateUnset .= "{$row->id},";
                        $this->retrievePbxNotifyPresence($presence, $row);
                        break;
                    }
                case self::calStatusNotify:     
                    $updateNotfiy .= "{$row->id},";
                    $presence->note = $this->getNote($row, "", true);
                    $presence->activity = $this->mapActivity($row->busy_type);
                    break;
                case self::calStatusStartSet:   
                    $updateStart .= "{$row->id},";
                    break;
            }
            $this->setPresence($presence, $row->guid, $row->h323);
        }
        $this->updatePbxStatus($updateUnset, self::calStatusNotify);
        $this->updatePbxStatus($updateNotfiy, self::calStatusStartSet);
        $this->updatePbxStatus($updateStart, self::calStatusEndSet);
    }
    
    private function retrievePbxNotifyPresence(PBX_Presence &$presence, &$row) {
        $activity = ""; // currently, $activity is only set for OOF!
        switch($row->busy_type) {
            case EWS_CalendarItemType::LegacyFreeBusyStatus_Busy: $activity = tl::tl(/* calendar state busy */ "Busy"); $activity = ""; break;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_Free: $activity = tl::tl(/* calendar state free */ "Free"); $activity = ""; break;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_OOF: $activity = tl::tl(/* calendar state out of office */ "Out of office"); break;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_WorkingElsewhere: $activity = tl::tl(/* calendar state working elsewhere */ "Working elsewhere"); break;
            default: $activity = tl::tl(/* calendar state unknown */ "Unknown"); $activity = ""; break;
        }
        $dateTime = $this->convertTime($row->time_start);
        $presence->note = tl::tlx(/* {$arg[1]} is a date with time and {$arg[2]} is a subject for a calendar entry */ 'Free until {$arg[1]} ({$arg[2]})', $dateTime, $this->getNote($row, $activity));
    }
    
    private function translateDate($d) {
        return tl::tlx( /* date format, args are 1: year, 2: month, 3: day */ '{$arg[2]}/{$arg[3]}/{$arg[1]}', date('Y', $d), date('m', $d), date('d', $d));
    }
    
    private function convertTime($sqlTime) {
        $time = strtotime($sqlTime);
        $dateTime = "";
        $timeStr = date("H:i", $time);
        if($timeStr == "00:00") {
            $timeTmp = strtotime("-1 day", $time);
            if(date("d") != date("d", $timeTmp)) {
                $timeStr = "";
                $time = $timeTmp;
            }
        }
        if(date("d") != date("d", $time)) {
            $dateTime = $this->translateDate($time);
            if($timeStr != "") {
                $dateTime .= " ";
            }
        }
        if($timeStr != "") {
            $dateTime .= $timeStr;
        }
        return $dateTime;
    }
    
    public function updatePbxDatePresence() {
        $exchangeDb = get_exchange_db();
        $query = "SELECT * FROM calendar_items ci
                    JOIN users u ON ci.user_id=u.id
                    WHERE ci.pbx_status=".self::calStatusNotify." AND u.server_id={$this->server->id}";
        $result = $exchangeDb->run($query);
        while(($row = pg_fetch_object($result))) {
            $presence = new PBX_Presence();
            $presence->note = null;
            $presence->activity = null;
            $presence->status = null;
            $this->retrievePbxNotifyPresence($presence, $row);
            $this->setPresence($presence, $row->guid, $row->h323);
        }
        $rows = pg_num_rows($result);
        $this->debug("{$rows} PBX presence messages updated");
    }
    
    private function getNote(&$row, $activity, $showEndTime = false) {
        $note = "";
        if($this->server->ews->omitSubject) {
            $note = tl::tl("Appointment");
        }
        else if($row->is_private == "t") {   // "t" == true...
            $note = tl::tl(/* A private calendar entry */ "Private");
        }
        else {
            if($row->subject == "") {
                $note = tl::tl(/* A calendar entry with empty subject. */ "Unknown");
            }
            else {
                $note = $row->subject;
            }
            if($row->location != "") {
                $note .= ", {$row->location}";
            }
        }
        if($activity != "") {
            $note = "$activity: $note";
        }
        if($showEndTime) {
            $note .= " ".tl::tl(/*until time/date*/"until")." ".$this->convertTime($row->time_end);
        }
        return $note;
    }
    
    private function mapActivity($busyType) {
        switch($busyType) {
            case EWS_CalendarItemType::LegacyFreeBusyStatus_Busy:
                return PBX_Presence::activity_busy;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_Free:       // ignore
                return PBX_Presence::activity_working;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_NoData:
                return PBX_Presence::activity_other;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_OOF:
                return PBX_Presence::activity_away;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_WorkingElsewhere:
                return PBX_Presence::activity_away;
            case EWS_CalendarItemType::LegacyFreeBusyStatus_Tentative:  // ignore
                return PBX_Presence::activity_unknown;
            default:
                return PBX_Presence::activity_unknown;
        }
    }
    
    private function updateLastRun() {
        $exchangeDb = get_exchange_db();
        $query = "UPDATE servers SET last_update=CAST(now() AS timestamp without time zone) WHERE id={$this->server->id};";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("Update of last run timestamp failed", pg_last_error($exchangeDb->db), Alarms::alExchangeLastRun);
        }
        clear_alarm(Alarms::alExchangeLastRun, ExchangeMethods::getAlarmServer(), "Exchange");
    }
    
    private function updatePbxStatus($ids, $newStatus) {
        $ids = rtrim($ids, ",");
        if($ids == "") return;
        $exchangeDb = get_exchange_db();
        $query = "UPDATE calendar_items SET pbx_status=$newStatus WHERE id IN ($ids)";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("Calendar item pbx_status update to $newStatus failed", pg_last_error($exchangeDb->db), Alarms::alExchangeUpdatePbxStatus);
        }
        clear_alarm(Alarms::alExchangeUpdatePbxStatus, ExchangeMethods::getAlarmServer(), "Exchange");
    }
    
    private function updateUser(PbxUser &$user, $h323, $emailStr, $smtp) {
        if($smtp != "" || $user->smtp != "") {
            $this->debug("User {$user->h323} ({$user->emails}|{$user->smtp}) changed to {$h323} ($emailStr|$smtp)");
        }
        $smtpStr = ($smtp == "" ? "NULL" : "'".pg_escape_string($smtp)."'");
        $exchangeDb = get_exchange_db();
        $query = "UPDATE users SET h323='".pg_escape_string($h323)."', emails='$emailStr', primary_smtp=$smtpStr WHERE id={$user->id}";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("User update failed", pg_last_error($exchangeDb->db), Alarms::alExchangeUpdateUser);
        }
        clear_alarm(Alarms::alExchangeUpdateUser, ExchangeMethods::getAlarmServer(), "Exchange");
        $user->h323 = $h323;
        $user->emails = $emailStr;
        $user->smtp = $smtp;
    }
    
    private function updateUserGuid(PbxUser &$user, $guid) {
        $exchangeDb = get_exchange_db();
        $query = "UPDATE users SET guid='".pg_escape_string($guid)."' WHERE id={$user->id}";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("User guid update failed", pg_last_error($exchangeDb->db), Alarms::alExchangeUpdateUserGuid);
        }
        clear_alarm(Alarms::alExchangeUpdateUserGuid, ExchangeMethods::getAlarmServer(), "Exchange");
        unset($this->users[$user->guid]);
        $user->guid = $guid;
        $this->users[$user->guid] = $user;
        
    }
    
    private function updateUserAvailability(PbxUser &$user) {
        $exchangeDb = get_exchange_db();
        $query = "UPDATE users SET availability_status=date_trunc('second', localtimestamp) WHERE id={$user->id}";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("User availability update failed", pg_last_error($exchangeDb->db), Alarms::alExchangeUpdateUserAvailability);
        }
        clear_alarm(Alarms::alExchangeUpdateUserAvailability, ExchangeMethods::getAlarmServer(), "Exchange");
    }
    
    private function updateUserSubscription(PbxUser &$user, $subscriptionId) {
        $exchangeDb = get_exchange_db();
        $subStr = (isset($subscriptionId) ? "'".pg_escape_string($subscriptionId)."'" : "NULL");
        $dateStr = (isset($subscriptionId) ? "date_trunc('second', localtimestamp)" : "NULL");
        $query = "UPDATE users SET subscription_id=$subStr, subscription_status=$dateStr, availability_status=NULL WHERE id={$user->id}";
        if($exchangeDb->run($query) === false) {
            throw new PbxException("User subscription update failed", pg_last_error($exchangeDb->db), Alarms::alExchangeUpdateUserSubscription);
        }
        clear_alarm(Alarms::alExchangeUpdateUserSubscription, ExchangeMethods::getAlarmServer(), "Exchange");
        if(isset($subscriptionId)) {
            $user->subscriptionId = $subscriptionId;
            $user->subscriptionStatus = time();
        }
        else {
            $user->subscriptionId = "";
            $user->subscriptionStatus = 0;
        }
        $user->availabilityStatus = 0;
    }
    
    private function deleteUsers($ids) {
        $exchangeDb = get_exchange_db();
        $query = "DELETE FROM calendar_items WHERE user_id IN ($ids)";
        $query2 = "DELETE FROM users WHERE id IN ($ids)";
        if($exchangeDb->run($query) === false || $exchangeDb->run($query2) === false) {
            throw new PbxException("User delete failed", pg_last_error($exchangeDb->db), Alarms::alExchangeDeleteUsers);
        }
        clear_alarm(Alarms::alExchangeDeleteUsers, ExchangeMethods::getAlarmServer(), "Exchange");
        $this->debug("User ids $ids deleted");
    }
    
    private function unsubscribe(PbxUser &$user) {
        if($user->subscriptionId == "") return;
        
        $user->subscriptionId = "";
        $user->subscriptionStatus = 0;
        $this->setPresence(new PBX_Presence(), $user->guid, $user->h323);
        $this->updateUserSubscription($user, null);
        $this->deleteCalendarItems($user);
        $this->debug("Unsubscribed user {$user->h323}");
    }
    
    private function subscribe(PbxUser &$user, $recursive = false) {
        $folderId = new EWS_DistinguishedFolderIdType();
        $folderId->Id = EWS_DistinguishedFolderIdType::Id_calendar;
        $folderId->Mailbox = new EWS_EmailAddressType();
        $folderId->Mailbox->EmailAddress = $user->smtp;
        $push = new EWS_PushSubscriptionRequestType();
        $push->FolderIds = new EWS_NonEmptyArrayOfBaseFolderIdsType();
        $push->FolderIds->DistinguishedFolderId = array();
        $push->FolderIds->DistinguishedFolderId[] = $folderId;
        $push->EventTypes = new EWS_NonEmptyArrayOfNotificationEventTypesType();
        $push->EventTypes->EventType[] = EWS_NonEmptyArrayOfNotificationEventTypesType::EventType_FreeBusyChangedEvent;
        $push->EventTypes->EventType[] = EWS_NonEmptyArrayOfNotificationEventTypesType::EventType_ModifiedEvent;
        $push->StatusFrequency = $this->server->ews->statusFrequency;
        $push->SubscribeToAllFolders = false;
        $push->URL = $this->linuxUrl;

        $sub = new EWS_SubscribeType();
        $sub->PushSubscriptionRequest = $push;
        $response = null;
        try
        {
            $response = $this->ews->Subscribe($sub, null, null, $this->version);
        }
        catch(Exception $e) {
            //throw new PbxException("User subscribe failed", $e->getMessage(), Alarms::alExchangeSubscribe);
            $response = null;
            $this->debug("ERROR: User subscribe failed: ".$e->getMessage());
        }
        clear_alarm(Alarms::alExchangeSubscribe, ExchangeMethods::getAlarmServer(), "Exchange");
        if($response != null && $response->ResponseMessages->SubscribeResponseMessage) {
            /** @var EWS_SubscribeResponseMessageType */ $test = $response->ResponseMessages->SubscribeResponseMessage;
            if($test->ResponseClass == EWS_SubscribeResponseMessageType::ResponseClass_Success && $test->SubscriptionId != "") {
                $this->debug("Subscription for {$user->smtp}: {$test->SubscriptionId}");
                $this->updateUserSubscription($user, $test->SubscriptionId);
            }
            else {
                if($test->ResponseCode == EWS_ResponseMessageType::ResponseCode_ErrorNonPrimarySmtpAddress && !$recursive) {
                    $emails = explode(";", $user->emails);
                    $error = "";
                    $smtp = $this->getPrimarySmtp($emails, $error);
                    if($smtp != "" && $smtp != $user->smtp) {
                        $this->updateUser($user, $user->h323, $user->emails, $smtp);
                        $this->subscribe($user, true);
                        return;
                    }
                    else if($smtp == "") {
                        $this->debug("subscribe getPrimatySmtp failed: $error");
                    }
                }
                $this->debug("Subscription for {$user->smtp} failed: ".$test->MessageText);
            }
        }
    }
    
    private function addUser(PbxUser &$user, $emails) {
        $error = "";
        $user->smtp = $this->getPrimarySmtp($emails, $error);
        $emailStr = implode(';', $emails);
        if($user->smtp == "") {
            $this->debug("addUser getPrimatySmtp failed ($emailStr): $error");
        }
        $exchangeDb = get_exchange_db();
        $smtp = ($user->smtp == "" ? "NULL" : "'".pg_escape_string($user->smtp)."'");
        $query = "INSERT INTO users (server_id,guid,h323,emails,primary_smtp) VALUES ({$this->server->id}, '".pg_escape_string($user->guid)."', '".pg_escape_string($user->h323)."', '".pg_escape_string($emailStr)."', $smtp) RETURNING id";
        $result = $exchangeDb->run($query);
        if($result === false) {
            throw new PbxException("User insert failed", pg_last_error($exchangeDb->db), Alarms::alExchangeInsertUser);
        }
        clear_alarm(Alarms::alExchangeInsertUser, ExchangeMethods::getAlarmServer(), "Exchange");
        $row = pg_fetch_object($result);
        $user->id = $row->id;
        $user->exists = true;
        $this->users[$user->guid] = $user;
    }
    
    private function getPrimarySmtp($emails, &$error) {
        $error = "";
        $prefixedEmails = array();
        foreach($emails as $email) {
            $prefixedEmails[] = "smtp:$email";  // search smtp addresses
            $prefixedEmails[] = "sip:$email";   // search sip addresses
            $prefixedEmails[] = "$email";       // search primary SMTP address
        }
        foreach($prefixedEmails as $email) {
            $request = new EWS_ResolveNamesType();
            //$request->ContactDataShape = EWS_ResolveNamesType::ContactDataShape_Default;
            $request->ReturnFullContactData = false;
            $request->UnresolvedEntry = "$email";
            try
            {
                /** @var EWS_ResolveNamesResponseType */$result = $this->ews->ResolveNames($request, null, null, $this->version);
                //var_dump($result->ResponseMessages->ResolveNamesResponseMessage);
                if($result->ResponseMessages && $result->ResponseMessages->ResolveNamesResponseMessage != null) {
                    /** @var EWS_ResolveNamesResponseMessageType */ $response = $result->ResponseMessages->ResolveNamesResponseMessage;
                    if($response->ResponseClass == EWS_ResponseMessageType::ResponseClass_Error) {
                        if(strpos($error, $response->ResponseCode) === false) {
                            $error .= $response->ResponseCode.": ".$response->MessageText.";";
                        }
                    }
                    if($response->ResolutionSet && $response->ResolutionSet->Resolution && $response->ResolutionSet->Resolution->Mailbox && $response->ResolutionSet->Resolution->Mailbox->EmailAddress != "") {
                        return $response->ResolutionSet->Resolution->Mailbox->EmailAddress;
                    }
                }
            }
            catch(Exception $e) {
                $this->debug("Resolving primary smtp address failed (".implode(",", $emails)."): ".$e->getMessage());
            }
        }
        clear_alarm(Alarms::alExchangeResolveNames, ExchangeMethods::getAlarmServer(), "Exchange");
        return "";
    }
}

function run_servers() {
    $exchangeDb = get_exchange_db();
    $query = "SELECT s.id 
                FROM servers s
                WHERE s.last_update IS NULL OR (s.last_update + CAST(s.interval || ' minute' AS INTERVAL)) < now()";
    $result = $exchangeDb->run($query);
    while(($row = pg_fetch_object($result))) {
        $ex = new ExchangeAvailability($row->id);
        $ex->run();
    }
}

function nightly_update() {
    $time = time();
    /* @var $exchangeSession ExchangeSession*/ global $exchangeSession;
    if(date("H", $time) == "00" && ($exchangeSession->exchangeCfg->lastPresenceUpdate + 60*60*24) <= $time) {  // just check this at night
        $exchangeDb = get_exchange_db();
        $query = "SELECT DISTINCT s.id FROM servers s
                    JOIN users u ON s.id=u.server_id
                    JOIN calendar_items ci ON ci.user_id=u.id
                    WHERE ci.pbx_status=".ExchangeAvailability::calStatusNotify;
        $result = $exchangeDb->run($query);
        while(($row = pg_fetch_object($result))) {
            $ex = new ExchangeAvailability($row->id);
            try
            {
                $ex->updatePbxDatePresence();
            }
            catch(PbxException $e) {
                $ex->handlePbxException($e);
            }
        }
        $exchangeSession->exchangeCfg->lastPresenceUpdate = strtotime(date("Y-m-d", $time));
        if($exchangeSession->exchangeCfg->writeConfig() === false) {
            echo ExchangeAvailability::debugPrefix()."Updating last presence date update time failed: ".$exchangeSession->exchangeCfg->error;
        }
    }
}

run_servers();
nightly_update();
close_exchange_db();
global $mainSession;
$mainSession->destroy();
?>